#include "rpi_leddriver.h"

#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "../common.h"

#include "../dma/rpi_dma.h"
#include "../dma/rpi_videocore.h"
#include "../gpio/rpi_gpio.h"
#include "../smi/rpi_smi.h"

#if PHYS_REG_BASE == PI_4_REG_BASE // Timings for RPi v4 (1.5 GHz)
#define SMI_TIMING 10, 15, 30, 15  // 400 ns cycle time
#else                              // Timings for RPi v0-3 (1 GHz)
#define SMI_TIMING 10, 10, 20, 10  // 400 ns cycle time
#endif

#define TX_TEST 0      // If non-zero, use dummy Tx data
#define LED_NBITS 24   // Number of data bits per LED
#define LED_PREBITS 4  // Number of zero bits before LED data
#define LED_POSTBITS 4 // Number of zero bits after LED data
#define BIT_NPULSES 3  // Number of O/P pulses per LED bit

// Length of data for 1 row (1 LED on each channel)
#define LED_DLEN (LED_NBITS * BIT_NPULSES)

// Transmit data type, 8 or 16 bits
#if LED_NCHANS > 8
#define TXDATA_T uint16_t
#else
#define TXDATA_T uint8_t
#endif

// Ofset into Tx data buffer, given LED number in chan
#define LED_TX_OSET(n) (LED_PREBITS + (LED_DLEN * (n)))

// Size of data buffers & NV memory, given number of LEDs per chan
#define TX_BUFF_LEN(n) (LED_TX_OSET(n) + LED_POSTBITS)
#define TX_BUFF_SIZE(n) (TX_BUFF_LEN(n) * sizeof(TXDATA_T))
#define VC_MEM_SIZE (PAGE_SIZE + TX_BUFF_SIZE(CHAN_MAXLEDS))

/* Global */
MEM_MAP vc_mem;
TXDATA_T *txdata;
TXDATA_T tx_buffer[TX_BUFF_LEN(CHAN_MAXLEDS)];

void swap_bytes();

void leddriver_setup() {
  gpio_setup();
  videocore_setup(&vc_mem, VC_MEM_SIZE);
  smi_setup(LED_NCHANS, SMI_TIMING, &vc_mem, TX_BUFF_LEN(CHAN_MAXLEDS), &txdata);
}

void leddriver_close() {
  printf("Closing\n");
  videocore_close(&vc_mem);
  smi_close(LED_NCHANS);
  gpio_close();
}

void set_color(uint32_t rgb, int index) {
  int msk;
  TXDATA_T *txd = &(tx_buffer[LED_TX_OSET(index)]);

  // For each bit of the 24-bit RGB values..
  for (size_t n = 0; n < LED_NBITS; n++) {
    // Mask to convert RGB to GRB, M.S bit first
    msk = n == 0 ? 0x800000 : n == 8 ? 0x8000 : n == 16 ? 0x80 : msk >> 1;
    // 1st byte or word is a high pulse on all lines
    txd[0] = (TXDATA_T)0xffff;
    // 2nd has high or low bits from data
    // 3rd is a low pulse
    txd[1] = txd[2] = 0;
    if (rgb & msk) {
      txd[1] = (TXDATA_T)0xffff;
    }
    txd += BIT_NPULSES;
  }
}

// Set Tx data for 8 or 16 chans, 1 LED per chan, given 1 RGB val per chan
// Logic 1 is 0.8us high, 0.4 us low, logic 0 is 0.4us high, 0.8us low
void rgb_txdata(int *rgbs, int index) {
  int i, n, msk;
  TXDATA_T *txd = &(tx_buffer[LED_TX_OSET(index)]);

  // For each bit of the 24-bit RGB values..
  for (n = 0; n < LED_NBITS; n++) {
    // Mask to convert RGB to GRB, M.S bit first
    msk = n == 0 ? 0x800000 : n == 8 ? 0x8000 : n == 16 ? 0x80 : msk >> 1;
    // 1st byte or word is a high pulse on all lines
    txd[0] = (TXDATA_T)0xffff;
    // 2nd has high or low bits from data
    // 3rd is a low pulse
    txd[1] = txd[2] = 0;
    for (i = 0; i < LED_NCHANS; i++) {
      if (rgbs[i] & msk)
        txd[1] |= (1 << i);
    }
    txd += BIT_NPULSES;
  }
}

void leddriver_refresh() {
#if LED_NCHANS <= 8
  swap_bytes();
#endif

  while (dma_active(DMA_CHAN_A)) {
    usleep(10);
  }

  memcpy(txdata, tx_buffer, TX_BUFF_SIZE(CHAN_MAXLEDS));
  enable_dma(DMA_CHAN_A);
  start_smi(&vc_mem);
  usleep(10);
}

// Swap adjacent bytes in transmit data
void swap_bytes() {
  uint16_t *wp = (uint16_t *)tx_buffer;
  int len = TX_BUFF_SIZE(CHAN_MAXLEDS);

  len = (len + 1) / 2;
  while (len-- > 0) {
    *wp = __builtin_bswap16(*wp);
    wp++;
  }
}