#include <ctype.h>
#include <pthread.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "drivers/leddriver/rpi_leddriver.h"
#include "drivers/selector/rpi_selector.h"

#include "rpi_midi_controller.h"
#include "tasks/artnet/rpi_artnet.h"
#include "tasks/cava/rpi_cava.h"

/*  Command-line parameters */
bool IsTestMode = false;
int chanLedCount = 0;

pthread_t *bgTasks[4] = {NULL};

void parseCommandLineArgs(int argc, char const *argv[]);

void terminate(int sig);

void execute_test_mode();

void execute_artnet_mode();

void execute_autonomous_mode();

void execute_manual_mode();

void adjust_loop(struct timespec const *loopStart);

int main(int argc, char const *argv[]) {
  int previousMode = -1;

  // setup
  parseCommandLineArgs(argc, argv);

  signal(SIGINT, terminate);

  struct sched_param sp;
  sp.sched_priority = 32;
  if (pthread_setschedparam(pthread_self(), SCHED_FIFO, &sp)) {
    fprintf(stderr, "WARNING: Failed to set stepper thread to real-time priority\n");
  }

  setup_selector();

  // // setup led
  // map_devices();
  // init_smi(LED_NCHANS, SMI_TIMING);
  // map_uncached_mem(&vc_mem, VC_MEM_SIZE);

  // setup_smi_dma(&vc_mem, TX_BUFF_LEN(chanLedCount), &txdata);

  leddriver_setup();

  setup_midi_controller();
  artnet_init();
  setup_cava();

  // loop
  while (1) {
    struct timespec loopStart;
    clock_gettime(CLOCK_MONOTONIC, &loopStart);
    int mode = get_selector_position();
    execute_midi_controller();

    if (mode != -1) {
      if (mode != previousMode) {
#if defined(_DEBUG)
        printf("swtching to mode : %d\n", mode);
#endif
        /* stop previous bg task */
        switch (previousMode) {
        case 1:
          stop_artnet_bg_worker();
          break;
        case 2:
          stop_cava_bg_worker();
          break;
        default:
          break;
        }

        /* start new bg task */
        switch (mode) {
        case 1:
          start_artnet_bg_worker();
          break;
        case 2:
          start_cava_bg_worker();
          break;
        default:
          break;
        }
      } else {
        switch (mode) {
        case 0:
          // mode test
          execute_test_mode();
          break;
        case 1:
          // artnet mode
          execute_artnet_mode();
          break;
        case 2:
          execute_autonomous_mode();
          break;
        case 3:
          // manual mode
          execute_manual_mode();
          break;

        default:
          if (mode != previousMode) {
            fprintf(stderr, "Error in selector\n");
          }
          break;
        }
      }
      previousMode = mode;
    }
    adjust_loop(&loopStart);
  }
}

void parseCommandLineArgs(int argc, char const *argv[]) {
  int args = 0;

  while (argc > ++args) // Process command-line args
  {
    if (argv[args][0] == '-') {
      switch (toupper(argv[args][1])) {
      case 'N': // -N: number of LEDs per channel
        if (args >= argc - 1)
          fprintf(stderr, "Error: no numeric value\n");
        else
          chanLedCount = atoi(argv[++args]);
        break;
      case 'T': // -T: test mode
        IsTestMode = true;
        break;
      default: // Otherwise error
        printf("Unrecognised option '%c'\n", argv[args][1]);
        printf("Options:\n"
               "  -n num    number of LEDs per channel\n"
               "  -t        Test mode (flash LEDs)\n");
        exit(1);
      }
    }
  }
}

void terminate(int sig) {
  leddriver_close();
  exit(0);
}

// Pointer to uncached Tx data buffer
// TXDATA_T tx_buffer[TX_BUFF_LEN(CHAN_MAXLEDS)]; // Tx buffer for assembling data

void execute_test_mode() {
  // RGB values for test mode (1 value for each of 16 channels)
  uint32_t on_rgbs[] = {0xef0000, 0x00ef00, 0x0000ef, 0xefef00, 0xef00ef, 0x00efef, 0xefefef};
  uint32_t off_rgbs = 0x000000;

  static int i = 0, offset = 0;

  for (size_t ledIndex = 0; ledIndex < chanLedCount; ++ledIndex) {
    set_color(ledIndex <= offset % chanLedCount ? on_rgbs[i] : off_rgbs, ledIndex);
  }

  leddriver_refresh();

  if (offset < chanLedCount) {
    ++offset;
  } else {
    offset = 0;
    if (i < 7) {
      ++i;
    } else {
      i = 0;
    }
  }
}

// RGB data
int rgb_data[CHAN_MAXLEDS][LED_NCHANS];

void execute_artnet_mode() {
  uint8_t *dmxData;

  for (size_t ledBar = 0; ledBar < LED_NCHANS; ledBar++) {
    if (artnet_get_dmx_data(ledBar, &dmxData) == 0) {
      for (size_t i = 0; i < chanLedCount; ++i) {
        uint8_t *rgb = dmxData + (i * 3);
        rgb_data[i][ledBar] = (rgb[0] << 16) | (rgb[1] << 8) | rgb[2];
        rgb_txdata(rgb_data[i], i);
      }
    }
  }

  leddriver_refresh();
}

void execute_autonomous_mode() {
  int ret;
  uint16_t *buffer;

  if ((ret = get_cava_buffer(&buffer)) == 0) {
    for (size_t ledBarIndex = 0; ledBarIndex < 4; ++ledBarIndex) {
      uint16_t barMax = 0;
      for (size_t bar = 0; bar < 20 / 4; ++bar) {
        unsigned barIndex = ledBarIndex * 20 / 4 + bar;
        if (barMax < buffer[barIndex]) {
          barMax = buffer[barIndex];
        }
      }
      unsigned ledToLight = barMax * chanLedCount / UINT16_MAX;

      for (size_t i = 0; i < ledToLight; ++i) {
        rgb_data[i][ledBarIndex] = 0xff0000;
        rgb_txdata(rgb_data[i], i);
      }
      for (size_t i = ledToLight; i < chanLedCount; ++i) {
        rgb_data[i][ledBarIndex] = 0x000000;
        rgb_txdata(rgb_data[i], i);
      }
    }

    leddriver_refresh();
  }
}

void execute_manual_mode() {}

void adjust_loop(struct timespec const *loopStart) {
  struct timespec loopEnd;
  long elapsedTimeUs, remainingTimeUs;

  clock_gettime(CLOCK_MONOTONIC, &loopEnd);
  elapsedTimeUs = (loopEnd.tv_nsec - loopStart->tv_nsec) / 1E3 +
                  ((unsigned)(loopEnd.tv_sec - loopStart->tv_sec)) * 1E6;
  remainingTimeUs = 16000 - elapsedTimeUs;

  if (remainingTimeUs >= 0) {
    // printf("loop remaining time %ld\n", remainingTimeUs);
    usleep(remainingTimeUs);
  } else {
    printf("loop overlap by %06ldus\n", -remainingTimeUs);
    printf("loop start %ld.%09lds\n", loopStart->tv_sec, loopStart->tv_nsec);
    printf("loop end %ld.%09lds\n", loopEnd.tv_sec, loopEnd.tv_nsec);
  }
}